// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit poa_impl;

interface

uses
  poa_int,orb_int,orb,types,classes,req_int,addr_int,map_int,env_int,
  iior_int,except_int,exceptions,policy_int,poa,code_int,OSThread,
  tcode,ir_int;


type

  TPOA_impl = class(TORBObject,IPOA,IObjectAdaptor)
  private
    FParent : IPOA;
    FUnique_id : _ulong;
    FDestructed: boolean;
    FManager : IPOAManager;
    FName : string;
    FOaid : ObjectID;
    FState : POAManagerState;
    FMediator : IPOAMEdiator;
    FEver_been_active: Boolean;
    FObjectMap : TObjectMap;
    FInvocationQueue : IInterfaceList;
    FChildList : TStrings;
    FDefaultServant : IServant;
    FServantManager : IServantManager;
    FAdapterActivator : IAdapterActivator;
    FObjectActivationLock: TMutex;
    //FCurrentKey: TThreadKey;
    {** serialization}
    FSerializeInvoke: TMutex;
    {** Fully Qualified Name }
    FFQN : string;
    // policies
    id_assignment_policy : IIdAssignmentPolicy;
    thread_policy : IThreadPolicy;
    lifespan_policy : ILifespanPolicy;
    id_uniqueness_policy : IIdUniquenessPolicy;
    implicit_activation_policy : IImplicitActivationPolicy;
    servant_retention_policy : IServantRetentionPolicy;
    request_processing_policy : IRequestProcessingPolicy;

    {** private object activation and deactivation}
    procedure remove_object(const id: ObjectID);
  protected
    function create_POA(adapter_name: string; a_POAManager: IPOAManager; policies: PolicyList): IPOA;
    function find_POA(adapter_name: string; activate_it: Boolean): IPOA;
    procedure _destroy(etherealize_objects: Boolean; wait_for_completion: Boolean);

    // factories for policies
    function create_thread_policy(value: ThreadPolicyValue): IThreadPolicy;
    function create_lifespan_policy(value: LifespanPolicyValue): ILifespanPolicy;
    function create_id_assignment_policy(value: IdAssignmentPolicyValue): IIdAssignmentPolicy;
    function create_id_uniqueness_policy(value: IdUniquenessPolicyValue): IIdUniquenessPolicy;
    function create_implicit_activation_policy(value: ImplicitActivationPolicyValue): IImplicitActivationPolicy;
    function create_servant_retention_policy(value: ServantRetentionPolicyValue): IServantRetentionPolicy;
    function create_request_processing_policy(value: RequestProcessingPolicyValue): IRequestProcessingPolicy;

    // Attributes
    function the_name: string;
    function the_parent: IPOA;
    function the_children: POAList;
    function the_POAManager: IPOAManager;
    function the_Activator: IAdapterActivator;
    procedure set_Activator(val: IAdapterActivator);

    // Servant Manager registration
    function get_servant_manager: IServantManager;
    procedure set_servant_manager(imgr: IServantManager);

    // operations for the USE_DEFAULT_SERVANT policy
    function get_servant: IServant;
    procedure set_servant(p_servant: IServant);

    // object activation and deactivation
    function activate_object(p_servant: IServant): ObjectID;
    procedure activate_object_with_id(id: ObjectID; p_servant: IServant);
    procedure deactivate_object(id: ObjectID);

    // reference creation operations
    function create_reference(intf: RepositoryID): IORBObject;
    function create_reference_with_id(id: ObjectID; intf: RepositoryID): IORBObject;

    // identity mapping operations
    function servant_to_id(p_servant: IServant): ObjectID;
    function servant_to_reference(p_servant: IServant): IORBObject;
    function reference_to_servant(reference: IORBObject): IServant;
    function reference_to_id(reference: IORBObject): ObjectID;
    function id_to_servant(id: ObjectID): IServant;
    function id_to_reference(id: ObjectID): IORBObject;

    procedure builtin_is_a(const ir: IInvocationRecord; const servant: IServant);
    procedure builtin_non_existent(const ir: IInvocationRecord; const servant: IServant);
    procedure builtin_interface(const ir: IInvocationRecord; const servant: IServant);
    
  protected
    // DORB extention
    procedure etherealize();
    function _find_POA(adapter_name: string; activate_it: Boolean): IPOA;
    procedure answer_invoke(id: ORBMessageID; const obj: IORBObject; const req: IORBRequest; stat: TInvokeStatus);
    function bind(id: ORBMessageID; const repoid: string; const oid: IObjectTag; const addr: IAddress): Boolean;
    function has_object(const obj: IORBObject): Boolean;
    function invoke(id: ORBMessageID;const obj : IORBObject; const req : IORBRequest; const pr: IPrincipal; response_exp: Boolean): Boolean;  stdcall;
    function is_local: Boolean;
    function skeleton(const obj : IORBObject): IORBObject;
    procedure shutdown(WaitForCompletion: Boolean);
    procedure cancel(id: ORBMessageID);
    function narrow_helper(const str: string): Pointer; override;
    procedure poa_manager_callback(state: POAManagerState;etherealize_objects: Boolean; wait_for_completion: Boolean);
    function ior_template(): IIOR;
    procedure local_invoke(const ir: IInvocationRecord);
    procedure perform_invoke(ir: IInvocationRecord);
    function builtin_invoke(const ir: IInvocationRecord; const servant: IServant): Boolean;
    procedure register_child(const cname : string; child: IPOA);
    procedure unregister_child(const cname : string);
    procedure set_policies(policies: PolicyList);
    function fqn: string;
    function get_oaid: ObjectID;
    function activate_for_this(p_servant: IServant): IORBObject;
    function _preinvoke(const obj: IORBObject): IServant;
    procedure _postinvoke();
    function ActiveObjectMap(): IActiveObjectMap;
  public
    {** Root_POA constructor}
    constructor Create(orb: IORB); overload;
    constructor Create(name: string; manager: IPOAManager; policies: PolicyList;
      parent: IPOA; orb: IORB); overload;
    destructor Destroy(); override;
    class function _narrow(const obj : IORBObject): IPOA;
  end;

implementation

uses
  address,windows,sysutils,map,throw,ior,req,static,poa_except,utils,
  policies,env,any;

var
  AllPOAs : TStrings;
  OAPrefix : ObjectID;
  idfactory : TUniqueIdGenerator;
  poauid : TUniqueIdGenerator;

procedure unregister_poa(const name: string);
var
  p: IPOA;
  i : integer;
begin
  i := ALLPOAs.IndexOf(name);
  if i >= 0 then
    begin
      p := IPOA(Pointer(ALLPOAs.Objects[i]));
      ALLPOAs.Delete(i);
      p._release;
    end;
end;

procedure register_poa(const name: string; const child: IPOA);
begin
  assert(ALLPOAs.IndexOf(name)<0);
  ALLPOAs.AddObject(name,Pointer(child));
  child._AddRef;
end;

//*********************************************************************
//  TPOA_impl
//*********************************************************************
{** constructor for creating root POA }
constructor TPOA_impl.Create(orb: IORB);
var
  pref: string;
  list: PolicyList;
  x : DWORD;
  t : TSystemTime;
begin
  FFQN := '';
  FObjectMap := TObjectMap.Create;
  FChildList := TStringList.Create;
  FORB := orb;
  FDestructed := false;
  FUnique_id := 0;
  FParent := nil;
  setLength(list,1);
  list[0] := create_implicit_activation_policy(IMPLICIT_ACTIVATION);
  set_policies(list);
  FManager := TPOAManager.Create;
  POACurrent := TPOACurrent.Create(orb);
  FInvocationQueue := TInterfaceList.Create;

  FDefaultServant := nil;
  FServantManager :=  nil;
  FAdapterActivator := nil;

  FName := 'RootPOA';
  FState := HOLDING;
  pref := '/' + IntToStr(GetCurrentProcessID());
  getSystemTime(t);
  with t do
    x := trunc(now())*MSecsPerDay + (wHour * 3600000 + wMinute * 60000 + wSecond * 1000 + wMilliseconds) ;
  pref := pref + '/' + IntToStr(x);
  OAPrefix := pref;
  FOaid := OAPrefix;
  FMediator := nil;
  FEver_been_active := false;
  _the_root_poa := self;
  register_poa(string(FOaid),self);
  FManager.add_managed_poa(self);
  FORB.set_initial_reference('RootPOA',self);
  FORB.register_oa(self,OAPriorityLocal);
  FObjectActivationLock := TMutex.Create;
  FSerializeInvoke := TMutex.Create;
end;

{** constructor for child POAs }
constructor TPOA_impl.Create(name: string; manager: IPOAManager;
  policies: PolicyList; parent: IPOA; orb: IORB);
var
  i : integer;
begin
  FName := name;
  FParent := parent;
  FORB := orb;
  FChildList := TStringList.Create;
  FObjectMap := TObjectMap.Create;
  FInvocationQueue := TInterfaceList.Create;

  set_policies(policies);

  if manager = nil then manager := TPOAManager.Create;
  FManager := manager;

  FDefaultServant := nil;
  FServantManager :=  nil;
  FAdapterActivator := nil;

  if parent.fqn <> '' then  FFQN := parent.fqn + '/';

  for i := 1 to Length(name) do
    if name[i] in ['/','\'] then   FFQN := FFQN + '\'
    else FFQN := FFQN + name[i];

  FState := manager.get_state;

  if lifespan_policy.value = TRANSIENT then
    FOaid := OAPrefix + '/' + poauid.new_id
  else
    assert(false,'not implemented');

  register_poa(FOaid,self);
  FManager.add_managed_poa(self);
  FParent.register_child(FName,self);
  FObjectActivationLock := TMutex.Create;
  FSerializeInvoke := TMutex.Create;
end;

function TPOA_impl.fqn: string;
begin
  result := FFQN
end;

procedure TPOA_impl.set_policies(policies: PolicyList);
var
  i : integer;
begin
  id_assignment_policy := nil;
  thread_policy := nil;
  lifespan_policy := nil;
  id_assignment_policy := nil;
  id_uniqueness_policy := nil;
  implicit_activation_policy := nil;
  servant_retention_policy := nil;
  request_processing_policy := nil;
  for i := 0 to Pred(length(policies)) do
    case policies[i].policy_type of
      THREAD_POLICY_ID : thread_policy := TThreadPolicy._narrow(policies[i]);
      LIFESPAN_POLICY_ID : lifespan_policy := TLifespanPolicy._narrow(policies[i]);
      ID_UNIQUENESS_POLICY_ID : id_uniqueness_policy := TIdUniquenessPolicy._narrow(policies[i]);
      ID_ASSIGNMENT_POLICY_ID : id_assignment_policy := TIdAssignmentPolicy._narrow(policies[i]);
      IMPLICIT_ACTIVATION_POLICY_ID : implicit_activation_policy := TImplicitActivationPolicy._narrow(policies[i]);
      SERVANT_RETENTION_POLICY_ID : servant_retention_policy := TServantRetentionPolicy._narrow(policies[i]);
      REQUEST_PROCESSING_POLICY_ID : request_processing_policy := TRequestProcessingPolicy._narrow(policies[i]);
    end;
  if thread_policy = nil then  thread_policy := create_thread_policy(ORB_CTRL_MODEL);
  if lifespan_policy = nil then lifespan_policy := create_lifespan_policy(TRANSIENT);
  if id_uniqueness_policy = nil then id_uniqueness_policy := create_id_uniqueness_policy(UNIQUE_ID);
  if implicit_activation_policy = nil then implicit_activation_policy := create_implicit_activation_policy(IMPLICIT_ACTIVATION);
  if servant_retention_policy = nil then servant_retention_policy := create_servant_retention_policy(RETAIN);
  if request_processing_policy = nil then request_processing_policy := create_request_processing_policy(USE_ACTIVE_OBJECT_MAP_ONLY);
  if id_assignment_policy = nil then id_assignment_policy := create_id_assignment_policy(SYSTEM_ID);
end;

procedure TPOA_impl.register_child(const cname: string; child: IPOA);
begin
  FChildList.AddObject(cname,TObject(Pointer(child)));
  child._addref;
end;

procedure TPOA_impl.unregister_child(const cname : string);
var
  i : integer;
  c : IPOA;
begin
  i := FChildList.IndexOf(cname);
  if i >= 0 then
    begin
      c := IPOA(Pointer(FChildList.objects[i]));
      FChildList.Delete(i);
      c._Release;
    end;
end;

procedure TPOA_impl.activate_object_with_id(id: ObjectID;
  p_servant: IServant);
var
  iface: string;
  por : IPOAObjectReference;
  al: IAuto_Lock;
begin
  if servant_retention_policy.value <> RETAIN then
    dorb_throw(WrongPolicy.Create());
  al := TAuto_Lock.Create(FObjectActivationLock);
  if FObjectMap.exists(id) then
    dorb_throw(ObjectAlreadyActive.Create);
  if (id_uniqueness_policy.value <> MULTIPLE_ID) and FObjectMap.exists(p_servant) then
    dorb_throw(ServantAlreadyActive.Create);
  // generate object reference
  iface := p_servant._primary_interface(id,self);
  por := TPoaObjectReference.Create(self,id,iface,p_servant);
  p_servant.activated_in(self);
  FObjectMap.add(por,p_servant);
end;

function TPOA_impl.create_id_assignment_policy(
  value: IdAssignmentPolicyValue): IIdAssignmentPolicy;
begin
  result := TIDAssignmentPolicy.Create(IMPLICIT_ACTIVATION_POLICY_ID,value)
end;

function TPOA_impl.create_implicit_activation_policy(
  value: ImplicitActivationPolicyValue): IImplicitActivationPolicy;
begin
  result := TImplicitActivationPolicy.Create(IMPLICIT_ACTIVATION_POLICY_ID,value)
end;


function TPOA_impl.create_id_uniqueness_policy(
  value: IdUniquenessPolicyValue): IIdUniquenessPolicy;
begin
  result := TIdUniquenessPolicy.Create(ID_UNIQUENESS_POLICY_ID,value)
end;

function TPOA_impl.create_servant_retention_policy(
  value: ServantRetentionPolicyValue): IServantRetentionPolicy;
begin
  result := TServantRetentionPolicy.Create(SERVANT_RETENTION_POLICY_ID,value)
end;

function TPOA_impl.create_lifespan_policy(
  value: LifespanPolicyValue): ILifespanPolicy;
begin
  result := TLifespanPolicy.Create(LIFESPAN_POLICY_ID,value)
end;

function TPOA_impl.create_POA(adapter_name: string;
  a_POAManager: IPOAManager; policies: PolicyList): IPOA;
var
  i : integer;
begin
  for i := 0 to Pred(FChildList.Count) do
    if IPOA(FChildList[i]).the_name = adapter_name then
      dorb_throw(AdapterAlreadyExists.Create);
  result := TPOA_impl.Create(adapter_name,a_POAManager,policies,self,FORB);
end;

function TPOA_impl.create_reference(intf: RepositoryID): IORBObject;
var
  uid, fullid: string;
  por : IPOAObjectReference;
begin
  if id_assignment_policy.value <> SYSTEM_ID then
    dorb_throw(WrongPolicy.Create);
  uid := string(idfactory.new_id());
  if lifespan_policy.value = PERSISTENT then
    fullid := string(oaprefix) + uid
  else
    fullid := uid;
  por := TPOAObjectReference.Create(self,fullid,intf);
  result := por.ref;
end;

function TPOA_impl.create_reference_with_id(id: ObjectID;
  intf: RepositoryID): IORBObject;
var
  por : IPOAObjectReference;
begin
  por := TPOAObjectReference.Create(self,id,intf);
  result := por.ref;
end;

function TPOA_impl.create_request_processing_policy(
  value: RequestProcessingPolicyValue): IRequestProcessingPolicy;
begin
  result := TRequestProcessingPolicy.Create(REQUEST_PROCESSING_POLICY_ID,value)
end;

function TPOA_impl.create_thread_policy(
  value: ThreadPolicyValue): IThreadPolicy;
begin
  result := TThreadPolicy.Create(THREAD_POLICY_ID,value)
end;

procedure TPOA_impl.deactivate_object(id: ObjectID);
var
  rec: IPOAObjectRecord;
  al: IAuto_Lock;
begin
  if servant_retention_policy.value <> RETAIN then
    dorb_throw(WrongPolicy.Create);
  al := TAuto_Lock.Create(FObjectActivationLock);
  if not FObjectMap.exists(id) then
    dorb_throw(ObjectNotActive.Create);

  rec := FObjectMap.find(id);
  rec.active := false;
  // there are no outstanding invocations, so remove it finaly
  if rec.invoke_cnt = 0 then
    remove_object(id);
end;

function TPOA_impl._find_POA(adapter_name: string;
  activate_it: Boolean): IPOA;
var
  i : integer;
begin
  result := nil;
  // search adapter
  for i := 0 to Pred(FChildList.Count) do
    if IPOA(FChildList[i]).the_name = adapter_name then
      begin
        result := IPOA(FChildList[i]);
        exit;
      end;
  if not activate_it or Not assigned(FAdapterActivator) then exit;
  if FState <> ACTIVE then exit;
  if not FAdapterActivator.unknown_adapter(self,adapter_name) then exit;
  // search again
  for i := 0 to Pred(FChildList.Count) do
    if IPOA(FChildList[i]).the_name = adapter_name then
      begin
        result := IPOA(FChildList[i]);
        exit;
      end;
end;

function TPOA_impl.find_POA(adapter_name: string;
  activate_it: Boolean): IPOA;
begin
  result := _find_POA(adapter_name,activate_it);
  if result = nil then
    dorb_throw(AdapterNonExistent.Create);
end;

function TPOA_impl.get_servant: IServant;
begin
  if request_processing_policy.value <> USE_DEFAULT_SERVANT then
    dorb_throw(WrongPolicy.Create);
  if FDefaultServant = nil then
    dorb_throw(NoServant.Create);
  result := FDefaultServant;
end;

function TPOA_impl.get_servant_manager: IServantManager;
begin
  if request_processing_policy.value <> USE_SERVANT_MANAGER then
    dorb_throw(WrongPolicy.Create);
  result := FServantManager
end;

function TPOA_impl.id_to_reference(id: ObjectID): IORBObject;
var
  rec: IPOAObjectRecord;
  al: IAuto_Lock;
begin
  if servant_retention_policy.value <> RETAIN then
    dorb_throw(WrongPolicy.Create);
  al := TAuto_Lock.Create(FObjectActivationLock);
  rec := FObjectMap.find(id);
  if rec = nil then
    dorb_throw(ObjectNotActive.Create);
  result := rec.por.ref;
end;

function TPOA_impl.id_to_servant(id: ObjectID): IServant;
var
  rec: IPOAObjectRecord;
  al: IAuto_Lock;
begin
  if servant_retention_policy.value <> RETAIN then
    dorb_throw(WrongPolicy.Create);
  al := TAuto_Lock.Create(FObjectActivationLock);
  rec := FObjectMap.find(id);
  if rec = nil then
    dorb_throw(ObjectNotActive.Create);
  result := rec.por.servant;
end;

function TPOA_impl.reference_to_id(reference: IORBObject): ObjectID;
var
  por : IPOAObjectReference;
begin
  por := TPOAObjectReference.Create(self,reference);
  if not por.is_legal() or not por.in_poa(PChar(FOaid)) then
    dorb_throw(WrongAdapter.Create());
  result := por.id;
end;

function TPOA_impl.reference_to_servant(reference: IORBObject): IServant;
var
  por : IPOAObjectReference;
begin
  por := TPOAObjectReference.Create(self,reference);
  if not por.is_legal() or not por.in_poa(PChar(FOaid)) then
    dorb_throw(WrongAdapter.Create());
  result := por.servant;
end;

function TPOA_impl.servant_to_id(p_servant: IServant): ObjectID;
var
  rec: IPOAObjectRecord;
  al: IAuto_Lock;
begin
  if (request_processing_policy.value = USE_DEFAULT_SERVANT) and
     (p_servant = FDefaultServant) and
     assigned(POACurrent) and
     POACurrent.iscurrent and
     (POACurrent.get_serv = p_servant) then
    begin
      result := POACurrent.get_object_id;
      exit;
    end;

  if servant_retention_policy.value <> RETAIN then
    dorb_throw(WrongPolicy.Create);

  al := TAuto_Lock.Create(FObjectActivationLock);
  if (implicit_activation_policy.value = IMPLICIT_ACTIVATION) and
    (id_uniqueness_policy.value = MULTIPLE_ID) then
      begin
        result := activate_object(p_servant);
        exit;
      end;

  rec := FObjectMap.find(p_servant);

  if rec <> nil then
    begin
      result := rec.por.id;
      exit;
    end;

  if (implicit_activation_policy.value = IMPLICIT_ACTIVATION) then
      begin
        result := activate_object(p_servant);
        exit;
      end;

  dorb_throw(ServantNotActive.Create);
end;

function TPOA_impl.servant_to_reference(p_servant: IServant): IORBObject;
var
  rec: IPOAObjectRecord;
  id: ObjectID;
  al: IAuto_Lock;
begin
  id := '';
  result := nil;

  if assigned(POACurrent) and
     POACurrent.iscurrent and
     (POACurrent.get_serv = p_servant) then
    begin
      result := POACurrent.make_ref;
      exit;
    end;

  if (servant_retention_policy.value <> RETAIN) or
     ((implicit_activation_policy.value <> IMPLICIT_ACTIVATION) and
     (id_uniqueness_policy.value <> UNIQUE_ID))  then
     dorb_throw(WrongPolicy.Create);

  al := TAuto_Lock.Create(FObjectActivationLock);
  rec := FObjectMap.find(p_servant);

  if (implicit_activation_policy.value = IMPLICIT_ACTIVATION) and
    ((id_uniqueness_policy.value = MULTIPLE_ID) or not assigned(rec)) then
     begin
       id := activate_object(p_servant);
       rec := FObjectMap.find(id);
     end;

  if assigned(rec) then
    begin
      result := rec.por.ref;
      Exit;
    end;

  dorb_throw(ServantNotActive.Create);

end;

procedure TPOA_impl.set_Activator(val: IAdapterActivator);
begin
  FAdapterActivator := val;
end;

procedure TPOA_impl.set_servant(p_servant: IServant);
begin
  FDefaultServant := p_servant;
end;

procedure TPOA_impl.set_servant_manager(imgr: IServantManager);
begin
  FServantManager := imgr;
end;

function TPOA_impl.the_Activator: IAdapterActivator;
begin
  result := FAdapterActivator
end;

function TPOA_impl.the_children: POAList;
var
  i : integer;
begin
  setLength(result,FChildList.count);
  for i := 0 to Pred(FChildList.count) do
    result[i] := IPOA(Pointer(FChildList.objects[i]));
end;

function TPOA_impl.the_name: string;
begin
  result := FName
end;

function TPOA_impl.the_parent: IPOA;
begin
  result := FParent
end;

destructor TPOA_impl.Destroy();
begin
  FObjectActivationLock.lock();
  try
    FObjectMap.Free;
  finally
    FObjectActivationLock.unlock();
  end; { try/finally }
  FChildList.free;
  FObjectActivationLock.Free;
  FSerializeInvoke.Free;
end;

procedure TPOA_impl.builtin_is_a(const ir: IInvocationRecord; const servant: IServant);
var
  req: IServerRequest;
  params : INVList;
  repoid : string;
  val : Boolean;
  a : IAny;
begin
  req := ir.make_dyn_req(self);
  params := TNVList.Create();
  params.add(ARG_IN);
  params.item(0).argument.set_type(_tc_string);
  req.arguments(params);
  params.item(0).argument.to_string(repoid);
  try
    val := servant._is_a(repoid);
  except
    on e: SystemException do
      begin
        req.set_exception(e);
        exit;
      end;
  end;
  a := TAny.Create;
  a.put_boolean(val);
  req.set_result(a);
end;

procedure TPOA_impl.builtin_non_existent(const ir: IInvocationRecord; const servant: IServant);
var
  req: IServerRequest;
  params : INVList;
  val : Boolean;
  a : IAny;
begin
  req := ir.make_dyn_req(self);
  params := TNVList.Create();
  req.arguments(params);
  try
    val := servant._non_existent();
  except
    on e: SystemException do
      begin
        req.set_exception(e);
        exit;
      end;
  end;
  a := TAny.Create;
  a.put_boolean(val);
  req.set_result(a);
end;

procedure TPOA_impl.builtin_interface(const ir: IInvocationRecord; const servant: IServant);
var
  req: IServerRequest;
  params : INVList;
  val : IInterfaceDef;
  a : IAny;
begin
  req := ir.make_dyn_req(self);
  params := TNVList.Create();
  req.arguments(params);
  try
    val := servant._get_interface();
  except
    on e: SystemException do
      begin
        req.set_exception(e);
        exit;
      end;
  end;
  if val = nil then
    begin
      req.set_exception(SystemException.Create_type(st_OBJ_ADAPTER,0,COMPLETED_NO));
      exit;
    end;
  a := TAny.Create;
  assert(false,'not implemented');
  req.set_result(a);
end;

function TPOA_impl.builtin_invoke(const ir: IInvocationRecord; const servant: IServant): Boolean;
var
  opname: string;
begin
  result := false;
  opname := ir.get_orequest.op_name;
  if strcomp(PChar(opname),'_is_a') = 0 then
    begin
      builtin_is_a(ir,servant);
      result := true;
    end
  else if strcomp(PChar(opname),'_non_existent') = 0 then
    begin
      builtin_non_existent(ir,servant);
      result := true;
    end
  else if strcomp(PChar(opname),'_interface') = 0 then
    begin
      builtin_interface(ir,servant);
      result := true;
    end;
end;

function TPOA_impl.ior_template(): IIOR;
begin
  result := FORB.ior_template();
end;

procedure TPOA_impl.poa_manager_callback(state: POAManagerState;etherealize_objects: Boolean; wait_for_completion: Boolean);
var
  rec: IInvocationRecord;
begin
  FState := state;

  if (state = ACTIVE) and not FEver_been_active and (FMediator <> nil) then
    begin
      FEver_been_active := true;
      // Mediator not implemented
    end;
  case state of
    HOLDING: { nothing to do};
    ACTIVE: while FInvocationQueue.Count > 0 do
        begin
          rec := FInvocationQueue[0] as IInvocationRecord;
          FInvocationQueue.Delete(0);
          rec.exec(self);
        end;
    INACTIVE: if etherealize_objects then
      etherealize;
    DISCARDING: while FInvocationQueue.Count > 0 do
        begin
          rec := FInvocationQueue[0] as IInvocationRecord;
          FInvocationQueue.Delete(0);
          rec.exec(self);
        end;
    else
      assert(false,'not implemented');
  end;
end;

function TPOA_impl.narrow_helper(const str: string): Pointer;
begin
  if strcomp(PChar(str),'IDL:omg.org/PortableServer/POA:1.0') = 0 then
    result := Pointer(self as IPOA)
  else
    result := nil;
end;

procedure TPOA_impl.perform_invoke(ir: IInvocationRecord);
var
  opname : string;
  rec : IPOAObjectRecord;
  servant : IServant;
  svr : IServerRequest;
  svrb : IServerRequestBase;
  ex: IORBException;
  sav : IServantActivator;
  ref: IPOAObjectReference;
  slo : IServantLocator;
  c : Cookie;
begin

  c := nil;
  opname := ir.get_orequest.op_name;

  FObjectActivationLock.lock();
  try
    if (servant_retention_policy.value = RETAIN) then
      begin
        rec := IPOAObjectRecord(FObjectMap.Find(ir.get_por));
        if assigned(rec) and rec.active then
          begin
            servant := rec.serv;
            assert(servant <> nil);
            ir.get_por.set_ref(rec.por.ref);
          end;
      end;
    // default servant
    if not assigned(servant) and (request_processing_policy.value = USE_DEFAULT_SERVANT) then
      begin
        if not assigned(FDefaultServant) then
          begin
            svr := ir.make_dyn_req(self);
            svr.set_exception(SystemException.Create_type(st_OBJ_ADAPTER,0,COMPLETED_NO));
            exit;
          end;
        servant := FDefaultServant;
      end;

    // servant adapter
    if not assigned(servant) and (request_processing_policy.value = USE_SERVANT_MANAGER) then
      begin
        if servant_retention_policy.value = RETAIN then
          begin
            sav := TServantActivator._narrow(FServantManager);
            if not assigned(sav) then
              begin
                svr := ir.make_dyn_req(self);
                svr.set_exception(SystemException.Create_type(st_OBJ_ADAPTER,0,COMPLETED_NO));
                exit;
              end;
            try
              servant := sav.incarnate(ir.get_por.id,self);
            except
              on e1: SystemException do
                begin
                  svr := ir.make_dyn_req(self);
                  svr.set_exception(e1);
                  exit;
                end;
              on e2: ForwardRequest do
                begin
                  FORB.answer_invoke(ir.id,InvokeForward,e2.forward_reference,ir.get_orequest);
                  exit;
                end;
            end;
            if not assigned(servant) then
              begin
                svr := ir.make_dyn_req(self);
                svr.set_exception(SystemException.Create_type(st_OBJ_ADAPTER,0,COMPLETED_NO));
                exit;
              end;
            // The new servant is placed in the active map
            ref := TPOAObjectReference.Create(ir.get_por);
            rec := FObjectMap.add(ref,servant);
          end
        else   // NON_RETAIN
          begin
            slo := TServantLocator._narrow(FServantManager);
            if not assigned(slo) then
              begin
                svr := ir.make_dyn_req(self);
                svr.set_exception(SystemException.Create_type(st_OBJ_ADAPTER,0,COMPLETED_NO));
                exit;
              end;
            try
              servant := slo.preinvoke(ir.get_por.id,self,ir.get_orequest.op_name,c);
            except
              on e1: SystemException do
                begin
                  svr := ir.make_dyn_req(self);
                  svr.set_exception(e1);
                  exit;
                end;
              on e2: ForwardRequest do
                begin
                  FORB.answer_invoke(ir.id,InvokeForward,e2.forward_reference,ir.get_orequest);
                  exit;
                end;
            end;
          end;
      end;


    if servant = nil then
      // cannot serve request
      begin
        svr := ir.make_dyn_req(self);
        ex := SystemException.Create_type(st_OBJECT_NOT_EXIST,0,COMPLETED_NO);
        svr.set_exception(ex);
        exit;
      end;
    // this make sure nobody is going to remove the object
    rec.invoke_cnt := Succ(rec.invoke_cnt);
  finally
    FObjectActivationLock.unlock();
  end; { try/finally }


  POACurrent._set(self,ir.get_por,servant);

  if not builtin_invoke(ir,servant) then
    begin
      svrb := ir.make_req(self, servant);
      // serialize invoke
      if thread_policy.value = SINGLE_THREAD_MODEL then begin
        FSerializeInvoke.lock();
        try
          servant.doinvoke(svrb);
        finally
          FSerializeInvoke.unlock();
        end; { try/finally }
      end
      else
        servant.doinvoke(svrb);
    end;

  POACurrent.unset();

  if assigned(slo) then
    slo.postinvoke(ir.get_por.id,self,ir.get_orequest.op_name,c,servant);

  FObjectActivationLock.lock();
  try
    rec.invoke_cnt := Pred(rec.invoke_cnt);
    if (rec.invoke_cnt = 0) and not rec.active then
      remove_object(ir.get_por.id);
  finally
    FObjectActivationLock.unlock();
  end; { try/finally }
end;

procedure TPOA_impl.local_invoke(const ir: IInvocationRecord);
var
  svr: IServerRequest;
  exc: IORBException;
begin
  // check for persistent POAs must be placed here...

  case FState of
    ACTIVE: ;
    HOLDING: begin
        FInvocationQueue.Add(ir);
        exit;
      end;
    DISCARDING: begin
        svr := ir.make_dyn_req(self);
        exc := SystemException.Create_type(st_TRANSIENT,0,COMPLETED_NO);
        svr.set_exception(exc);
        exit;
      end;
    INACTIVE: assert(false,'not implemented');
    else
      assert(false,'not implemented');
  end;

  // if request is not for us but for descendant, try to call adaptor activator
  if not ir.get_por.in_poa(PChar(FOaid)) then
    assert(false,'not implemented');
  perform_invoke(ir);
end;

function TPOA_impl.activate_object(p_servant: IServant): ObjectID;
var
  iface : string;
  uid,fulluid: ObjectID;
  por : IPOAObjectReference;
  al: IAuto_Lock;
begin
  al := TAuto_Lock.Create(FObjectActivationLock);
  assert(p_servant<>nil);

  if (id_assignment_policy.value <> SYSTEM_ID) or
     (servant_retention_policy.value <> RETAIN) then
    dorb_throw(WrongPolicy.Create);

  if (id_uniqueness_policy.value <> MULTIPLE_ID) and
     FObjectMap.exists(p_servant) then
    dorb_throw(ServantAlreadyActive.Create);

  uid := idfactory.new_id;

  if lifespan_policy.value = PERSISTENT then
    fulluid := OAPrefix + uid
  else
    fulluid := uid;

  iface := p_servant._primary_interface(fulluid,self);
  por := TPOAObjectReference.Create(self,fulluid,PChar(iface),p_servant);
  p_servant.activated_in(self);
  FObjectMap.add(por,p_servant);
  result := fulluid;
end;

class function TPOA_impl._narrow(const obj : IORBObject): IPOA;
var
  ptr: Pointer;
begin
  result := nil;
  if (obj <> nil) then
    begin
      ptr := obj.narrow_helper('IDL:omg.org/PortableServer/POA:1.0');
      if ptr <> nil then
        result := IPOA(ptr);
    end;
end;

function TPOA_impl.get_oaid(): ObjectID;
begin
  result := FOaid;
end;

procedure TPOA_impl._destroy(etherealize_objects, wait_for_completion: Boolean);
var
  rec: IInvocationRecord;
begin
  if FDestructed then exit;
  FDestructed := true;
  unregister_poa(string(FOaid));
  if Assigned(FParent) then
    FParent.unregister_child(FName);
  FManager.del_managed_poa(self);

  // POAMediator not implemented

  if not assigned(FParent) then //rootPOA
    FORB.unregister_oa(self);

  FState := INACTIVE;

  while FInvocationQueue.count > 0 do
    begin
      rec := FInvocationQueue[0] as IInvocationRecord;
      rec.exec(self);
      FInvocationQueue.Delete(0);
    end;

  if etherealize_objects then
    etherealize();
end;

function TPOA_impl.the_POAManager: IPOAManager;
begin
  result := FManager;
end;

procedure TPOA_impl.answer_invoke(id: ORBMessageID; const obj: IORBObject;
  const req: IORBRequest; stat: TInvokeStatus);
begin
  FORB.answer_invoke(id,stat,nil,req);
end;

function TPOA_impl.bind(id: ORBMessageID; const repoid: string;
  const oid: IObjectTag; const addr: IAddress): Boolean;
var
  i : integer;
  rec : IPOAObjectRecord;
  pif : string;
begin
  result := false;
  if (addr <> nil) and (addr.is_local) then exit;
  if servant_retention_policy.value = RETAIN then begin
    FObjectActivationLock.lock();
    try
      for i := 0 to Pred(FObjectMap.Count) do
        begin
          rec := FObjectMap.item(i);
          if not rec.serv._is_a(repoid) then
            begin
              pif := rec.serv._primary_interface(rec.por.id,self);
              if strcomp(PChar(repoid),PChar(pif)) <> 0 then
                continue;
            end;
          if (oid.get_length = 0) or (strlcomp(PChar(oid.get_bytes),PChar(rec.por.id),oid.get_length)=0) then
            begin
              FORB.answer_bind(id,LocateHere,rec.por.ref);
              result := true;
              exit;
            end;
        end;
    finally
      FObjectActivationLock.unlock();
    end; { try/finally }
  end;
  // ask children ...
  result := true;
  for i := 0 to pred(FChildList.Count) do
    if (IPOA(Pointer(FChildList.Objects[i]))).bind(id,repoid,oid,addr) then
      exit;
  result := false;
end;

procedure TPOA_impl.cancel(id: ORBMessageID);
var
  rec: IInvocationRecord;
  i : integer;
begin
  for i := 0 to Pred(FInvocationQueue.Count) do
    begin
      rec := FInvocationQueue[i] as IInvocationRecord;
      if ORB_Instance.get_msgid(rec.id) = ORB_Instance.get_msgid(id) then
        FInvocationQueue.Remove(rec);
    end;
  for i := 0 to Pred(FChildList.Count) do
    IPOA(Pointer(FChildList.objects[i])).cancel(id);
end;

function TPOA_impl.has_object(const obj: IORBObject): Boolean;
var
  l: _ulong;
  key: string;
begin
  result := false;
  if (obj = nil) or (obj.IOR = nil) or (obj.IOR.get_profile(TAG_ANY) = nil) then
    exit;
  key := obj.IOR.get_profile(TAG_ANY).objkey(l);
  if (key = '') or (l < 1) then exit;
  if (Length(OAPrefix) > 0) and (l >= _ulong(Length(OAPrefix))) then
    begin
      if strlcomp(PChar(key),PChar(OAPrefix),Length(OAPrefix))=0 then
        begin
          result := true;
          exit;
        end;
    end;
end;

function TPOA_impl.invoke(id: ORBMessageID; const obj: IORBObject;
  const req: IORBRequest; const pr: IPrincipal; response_exp: Boolean): Boolean;
var
  por: IPOAObjectReference;
  i : integer;
  p,iter : IPOA;
  exc : ISystemException;
  ir : IInvocationRecord;
  svr: IServerRequest;
  cname: PChar;
begin
  result := false;
  assert(IPOA(self) = _the_root_poa);
  por := TPOAObjectReference.Create(self,obj);
  assert(por.is_legal);

  if FDestructed then
    begin
      ir := TInvocationRecord.Create(id,por,req,pr);
      FInvocationQueue.Add(ir);
      exit;
    end;

  p := nil;
  i := ALLPOAs.IndexOf(por.poa_name);
  if i >= 0 then
    p := IPOA(Pointer(ALLPOAs.Objects[i]))
  else
    begin
      // only for persistent objects
      if (poa_impl_name = '') or (por.poa_name = '/') then
        begin
          ir := TInvocationRecord.Create(id,por,req,pr);
          svr := ir.make_dyn_req(self);
          exc := SystemException.Create_type(st_OBJECT_NOT_EXIST,0,COMPLETED_NO);
          svr.set_exception(exc);
          result := true;
          exit;
        end;
      // search in child POAs
      p := self;
      while true do
        begin
          if por.in_poa(PChar(p.get_oaid())) then break;
          cname := por.next_descendant_poa(PChar(p.fqn),PChar(poa_impl_name));
          try
            iter := p.find_POA(cname,false);
          except
            on e: SystemException do
              begin
                ir := TInvocationRecord.Create(id,por,req,pr);
                svr := ir.make_dyn_req(self);
                exc := SystemException.Create_type(st_OBJECT_NOT_EXIST,0,COMPLETED_NO);
                svr.set_exception(exc);
                result := true;
                exit;
              end;
          end;
          if iter = nil then
            break;
          p := iter;
        end;
    end;
  // found
  ir := TInvocationRecord.Create(id,por,req,pr);
  p.local_invoke(ir);
  result := true;
  exit;
end;

function TPOA_impl.is_local: Boolean;
begin
  result := true;
end;

procedure TPOA_impl.shutdown(WaitForCompletion: Boolean);
begin
assert(false,'not implemented');
end;

{**
 * skeleton() calls the servant's generated _make_stub method which
 * returns a POA-specific skeleton that invokes the servant directly
 * after asking the POA whether the invocation may proceed
}
function TPOA_impl.skeleton(const obj: IORBObject): IORBObject;
var
  por: IPOAObjectReference;
  rec: IPOAObjectRecord;
  poa: IPOA;
  i: integer;
begin
  result := nil;
  assert(IPOA(self) = _the_root_poa);
  por := TPOAObjectReference.Create(Self, obj);
  Assert(por.is_legal());
  i := ALLPOAs.IndexOf(por.poa_name);
  if i >= 0 then begin
    poa := IPOA(Pointer(ALLPOAs.Objects[i]));
    FObjectActivationLock.lock();
    try
      rec := poa.ActiveObjectMap.find(Self, obj);
      if rec <> nil then
        result := rec.serv._make_stub(Self, obj);
    finally
      FObjectActivationLock.unlock();
    end; { try/finally }
  end;
end;

procedure TPOA_impl.etherealize();
var
  sav: IServantActivator;
  rec : IPOAObjectRecord;
  serv: IServant;
  por: IPOAObjectReference;
  other : Boolean;
begin
  if (servant_retention_policy.value = RETAIN) and
     (request_processing_policy.value = USE_SERVANT_MANAGER) then
    sav := TServantActivator._narrow(FServantManager);

  FObjectActivationLock.lock();
  try
    while FObjectMap.count > 0 do
      begin
        rec := FObjectMap.pop;
        if sav <> nil then
          begin
            serv := rec.serv;
            por := rec.por;
            other := FObjectMap.exists(serv);
            try
              sav.etherealize(por.id,self,serv,true,other);
            except
            end;
          end;
      end;
  finally
    FObjectActivationLock.unlock();
  end; { try/finally }
end;

function TPOA_impl.activate_for_this(p_servant: IServant): IORBObject;
var
  rec: IPOAObjectRecord;
  id : ObjectID;
  al: IAuto_Lock;
begin
  if (servant_retention_policy.value <> RETAIN) or
     (id_uniqueness_policy.value <> UNIQUE_ID) then
    dorb_throw(WrongPolicy.Create());
  al := TAuto_Lock.Create(FObjectActivationLock);
  rec := FObjectMap.find(p_servant);
  if rec = nil then
    begin
      if implicit_activation_policy.value <> IMPLICIT_ACTIVATION then
        dorb_throw(WrongPolicy.Create());
      id := activate_object(p_servant);
      rec := FObjectMap.find(id);
      assert(rec<>nil);
    end;
  result := rec.por.ref
end;

procedure TPOA_impl._postinvoke;
begin
  POACurrent.unset;
end;

function TPOA_impl._preinvoke(const obj: IORBObject): IServant;
var
  rec : IPOAObjectRecord;
  por: IPOAObjectReference;
  serv: IServant;
  al: IAuto_Lock;
begin
  al := TAuto_Lock.Create(FObjectActivationLock);
  result := nil;
  case FState of
    HOLDING,DISCARDING: dorb_throw(SystemException.Create_type(st_TRANSIENT));
    INACTIVE : exit;
  end;
  rec := FObjectMap.find(self,obj);
  if assigned(rec) and rec.active then
    begin
      serv := rec.serv;
      por := rec.por;
    end
  else if (request_processing_policy.value = USE_DEFAULT_SERVANT) and assigned(FDefaultServant) then
    begin
      serv := FDefaultServant;
      por := TPOAObjectReference.Create(self,obj);
    end
  else
    exit;
  POACurrent._set(self,por,serv);
  result := serv;
end;

{** this methode has always to be called with the ObjectActivationLock held}
procedure TPOA_impl.remove_object(const id: ObjectID);
var
  rec: IPOAObjectRecord;
  delref: PTDeletionRecord;
  other: boolean;
  sav: IServantActivator;
  serv: IServant;
  por: IPOAObjectReference;
begin
  // we made a mistake somewhere is this should ever happen
  Assert(FObjectMap.exists(id));
  rec := FObjectMap.del(id);
  delref := rec.delref;
  other := FObjectMap.exists(rec.serv());
  // Etherealize object
  if FServantManager <> nil then begin
    sav := TServantActivator._narrow(FServantManager);
    Assert(sav <> nil);
    serv := rec.serv();
    por := rec.por();
    rec := nil;
    sav.etherealize(por.id(), Self, serv, false, other);
    por := nil;
  end
  else
    rec := nil;
  if delref <> nil then begin
    delref.signal := 1;
    delref.condition.broadcast();
  end;
end;

function TPOA_impl.ActiveObjectMap: IActiveObjectMap;
begin
  result := FObjectMap;
  result._AddRef;
end;

initialization
  poa_impl_name:= '';
  AllPOAs := TStringList.Create;
  idfactory := TUniqueIdGenerator.Create('_');
  poauid := TUniqueIdGenerator.Create('');
finalization
  AllPOAs.free;
  idfactory.free;
  poauid.free;
end.
